/*
 * Copyright (c) 2006-2007 Hypertriton, Inc. <http://www.hypertriton.com/>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/wait.h>

#include <EXTERN.h>
#include <perl.h>

#include "mailprocd.h"

#include <stdlib.h>
#include <limits.h>
#include <syslog.h>
#include <ctype.h>
#include <pwd.h>
#include <grp.h>

void
SF_MessageInit(SF_Message *msg)
{
	msg->svParsed = NULL;
	msg->svRewrite = NULL;
	msg->text = NULL;
	msg->text_len = 0;
	msg->status = NULL;
	msg->mail_from[0] = '\0';
	msg->rcpt_to[0] = '\0';
	msg->user_part[0] = '\0';
	msg->domain_part[0] = '\0';
	msg->ml = NULL;
}

void
SF_MessageInitStatus(SF_MsgStatus *st)
{
	st->score = 0.0;
	st->req_score = 100.0;
	st->spam_status = 0;
	st->sv = NULL;
}

SF_Message *
SF_MessageNew(void)
{
	SF_Message *msg;
	
	msg = Malloc(sizeof(SF_Message));
	SF_MessageInit(msg);
	return (msg);
}

void
SF_MessageFree(SF_Message *msg)
{
	if (msg->svParsed != NULL) { sv_free(msg->svParsed); }
	if (msg->status != NULL) {
		sv_free(msg->status->sv);
		free(msg->status);
	}
	if (msg->svRewrite != NULL) {
		sv_free(msg->svRewrite);
		/* msg->text was pointer to svRewrite */
	} else {
		free(msg->text);
	}
	if (msg->ml != NULL)
		free(msg->ml);
}

/*
 * Scan a ruleset for "spam" conditions to determine if we'll need
 * a spam check for classification.
 */
static int
SF_NeedFiltering(SF_Message *msg, SF_Ruleset *ruleset, int lvl)
{
	SF_Rule *rule;

	TAILQ_FOREACH(rule, &ruleset->rules, rules) {
		if (rule->flags & RULE_SMTP) {
			continue;
		}
		if (strncmp(rule->cond, "spam", 4) == 0) {
			return (1);
		}
		if (rule->insn[0] == '&') {
			SF_Ruleset *pRuleset;
			int rv;

			if (lvl >= RULE_MAX_RECURSION_LVL) {
				continue;
			}
			if ((pRuleset = LOCAL_GetRulesetByName(&rule->insn[1]))
			     == NULL) {
				continue;
			}
			rv = SF_NeedFiltering(msg, pRuleset, lvl+1);
			SF_RulesetFree(pRuleset);
			if (rv == 1 || rv == -1) {
				return (rv);
			}
			return (1);
		}
	}
	return (0);
}

static __inline__ int
SF_SpamTest(SF_Message *msg)
{
	char wbuf[32];
	int pp[2], status;
	pid_t pid;
	uid_t uid;
	gid_t gid;
	char *ep;
//	size_t i, len;
	size_t len;
	ssize_t rv;
	size_t nread;

	if (msg->text_len >= SPAM_TEST_SIZE_LIMIT) {
		msg->status->score = 0.0;
		msg->status->req_score = 123.0;
		Debug("Message over limit for spam test (%lu bytes)",
		    (unsigned long)msg->text_len);
		return (0);
	}
	if (LOCAL_GetDefaultRecipientUID(msg->rcpt_to, &uid, &gid) == -1) {
		SF_SetError("Cannot get default UID for %s", msg->rcpt_to);
		return (-1);
	}
	if (pipe(pp) == -1) {
		SF_SetError("pipe: %s", strerror(errno));
		return (-1);
	}
	if ((pid = fork()) == 0) {				/* Child */
		struct sigaction sa;
		sigset_t allsigs;
		size_t wrote;

		setproctitle("spam-test");

		sa.sa_handler = SIG_DFL;
		sa.sa_flags = 0;
		sigemptyset(&sa.sa_mask);
		sigaction(SIGPIPE, &sa, NULL);

		close(pp[0]);
		close(STDIN_FILENO);
		if (dup2(pp[1], STDOUT_FILENO) == -1) {
			SF_SetError("dup2: %s", strerror(errno));
			goto chld_fail;
		}
		
		/* Drop privileges irrevocably. */
		sigfillset(&allsigs);
		sigprocmask(SIG_BLOCK, &allsigs, NULL);
		if (setegid(gid) < 0 || setgid(gid) < 0 ||
		    seteuid(uid) < 0 || setuid(uid) < 0) {
			sigprocmask(SIG_UNBLOCK, &allsigs, NULL);
			SF_SetError("set*id(%d,%d): %s", (int)uid, (int)gid,
			    strerror(errno));
			goto chld_fail;
		}
		sigprocmask(SIG_UNBLOCK, &allsigs, NULL);

		/* Make damn sure we are not running as root. */
		if (geteuid() == 0 || getuid() == 0) {
			SF_SetError("Failed to drop privileges!");
			goto chld_fail;
		}

		/* Perform the SpamAssassin check. */
		if (SA_ParseMessage(msg) == -1) {
			SF_SetError("SA_ParseMessage");
			goto chld_fail;
		}
		if (SA_CheckMessage(msg) == -1) {
			SF_SetError("SA_ParseMessage");
			goto chld_fail;
		}
#if 1
		/* XXX TODO implement header insertion ourselves */
		if (SA_RewriteMessage(msg) == -1) {
			SF_SetError("SA_RewriteMessage");
			goto chld_fail;
		}
#endif
		/* Write the test score to the parent. */
		snprintf(wbuf, sizeof(wbuf), "%f", msg->status->score);
		if (write(STDOUT_FILENO, wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
			SF_SetError("Write error (score)");
			goto chld_fail;
		}
		
		/* Write the required score to the parent. */
		snprintf(wbuf, sizeof(wbuf), "%f", msg->status->req_score);
		if (write(STDOUT_FILENO, wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
			SF_SetError("Write error (req_score)");
			goto chld_fail;
		}

		/* Write the rewritten message back to the parent. */
		snprintf(wbuf, sizeof(wbuf), "%lu",
		    (unsigned long)msg->text_len);
		if (write(STDOUT_FILENO, wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
			SF_SetError("Write error (len)");
			goto chld_fail;
		}

		for (wrote = 0; wrote < msg->text_len;) {
			rv = write(STDOUT_FILENO, &msg->text[wrote],
			    msg->text_len - wrote);
			if (rv == -1) {
				if (errno == EINTR) {
					continue;
				}
				SF_SetError("Write error: %s", strerror(errno));
				goto chld_fail;
			} else if (rv == 0) {
				SF_SetError("Parent EOF");
				goto chld_fail;
			}
			wrote += rv;
		}
		exit(0);
chld_fail:
		syslog(LOG_ERR, "spam-test process: %s", SF_GetError());
		exit(1);
	} else if (pid == -1) {
		SF_SetError("fork: %s", strerror(errno));
		return (-1);
	}

	close(pp[1]);
	
	/* Read the test score. */
	if (read(pp[0], wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
		SF_SetError("Read error (score)");
		goto fail;
	}
	errno = 0;
	msg->status->score = (float)strtod(wbuf, &ep);
	if (*ep != '\0' || errno == ERANGE) {
		SF_SetError("Bad score: %s", wbuf);
		goto fail;
	}

	/* Read the required score. */
	if (read(pp[0], wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
		SF_SetError("Bad req_score value");
		goto fail;
	}
	errno = 0;
	msg->status->req_score = (float)strtod(wbuf, &ep);
	if (*ep != '\0' || errno == ERANGE) {
		SF_SetError("Bad req_score: %s", wbuf);
		goto fail;
	}

	/* Read the rewritten message from the child process. */
	if (read(pp[0], wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
		SF_SetError("Bad len value");
		goto fail;
	}
	errno = 0;
	len = (size_t)strtoul(wbuf, &ep, 10);
	if (*ep != '\0' || errno == ERANGE) {
		SF_SetError("Bad len: %s", wbuf);
		goto fail;
	}
	msg->text_len = len;
	if ((msg->text = realloc(msg->text, len+1)) == NULL) {
		SF_SetError("Out of memory for rewritten message (%lu)",
		    (unsigned long)msg->text_len);
		msg->text_len = 0;
		goto fail;
	}
	for (nread = 0; nread < len;) {
		rv = read(pp[0], &msg->text[nread], len - nread);
		if (rv == -1) {
			if (errno == EINTR) {
				continue;
			}
			SF_SetError("Read error: %s", strerror(errno));
			goto fail;
		} else if (rv == 0) {
			syslog(LOG_ERR, "EOF from spam-test process");
			break;
		}
		nread += rv;
	}
	msg->text[nread] = '\0';

#if 0
	/* XXX Check for abnormal NULs */
	for (i = 0; i < nread; i++) {
		if (msg->text[i] == '\0') {
			msg->text[i] = ' ';
			Debug("NUL at %lu in spam-test result!",
			    (unsigned long)i);
		}
	}
#endif

	Debug("Spam score = %f (%f required), %lu bytes",
	    msg->status->score, msg->status->req_score,
	    (unsigned long)msg->text_len);
wait:
	if (waitpid(pid, &status, 0) == -1) {
		if (errno == EINTR) {
			goto wait;
		}
		SF_SetError("wait: %s", strerror(errno));
		goto fail;
	}
	if (WIFSIGNALED(status)) {
		SF_SetError("signal %d", WTERMSIG(status));
		goto fail;
	} else if (WIFEXITED(status) && WEXITSTATUS(status)) {
		SF_SetError("failed (%d)", WEXITSTATUS(status));
		goto fail;
	}
	close(pp[0]);
	return (0);
fail:
	close(pp[0]);
	return (-1);
}

/* Main message delivery routine. */
void
SF_MessageProcess(SF_Message *msg)
{
	SF_Ruleset *ruleset;

	if ((ruleset = LOCAL_GetRulesetByRcpt(msg->rcpt_to)) == NULL) {
		SF_SetError("No ruleset for %s", msg->rcpt_to);
		goto fail_temp;
	} 
	if ((msg->status = malloc(sizeof(SF_MsgStatus))) == NULL) {
		SF_SetError("Out of memory");
		goto fail_temp;
	}
	SF_MessageInitStatus(msg->status);

	if (SF_NeedFiltering(msg, ruleset, 0)) {
		if (SF_SpamTest(msg) == -1)
			syslog(LOG_ERR, "Spam test failed: %s", SF_GetError());
	}

	/*
	 * Deliver the message according to user classification rulesets.
	 * The function will fail only if the ruleset cannot be looked up
	 * or contains errors.
	 *
	 * If actual delivery fails, the function will return success but
	 * add the message to the internal queue.
	 */
	if (SF_MessageClassify(msg, ruleset, 0) == -1) {
		SF_SetError("Classification failed: %s", SF_GetError());
		goto fail_temp;
	}
	SF_RulesetFree(ruleset);
	return;
fail_temp:
	if (ruleset != NULL) {
		SF_RulesetFree(ruleset);
	}
	syslog(LOG_ERR, "%s; reinjecting", SF_GetError());
	Debug("%s; reinjecting", SF_GetError());
	sleep(3);
	if (SMTP_ForwardMessage(msg, sfReinHost, sfReinPort, msg->rcpt_to)
	    == -1) {
		syslog(LOG_ERR, "Reinjection failed (%s); queueing!",
		    SF_GetError());
		/* TODO QUEUE */
	}
}

static __inline__ int
SF_MatchRule(SF_Message *msg, SF_Rule *rule)
{
	if (strcmp(rule->cond, "any") == 0) {
		return (1);
	}
	if (strncmp(rule->cond, "spam", 4) == 0) {
		float score;
		char *ep;

		if (rule->cond[4] == '\0' || rule->cond[5] == '\0' ||
		    rule->cond[6] == '\0') {
			return (1);
		}
		score = strtod(&rule->cond[6], &ep);
		if (*ep != '\0') {
			syslog(LOG_ERR, "<%s>: Bad rule syntax: `%s'",
			    msg->rcpt_to, rule->cond);
			return (0);
		}
		return (rule->cond[4] == '<' ? (msg->status->score <= score):
		                               (msg->status->score >= score));
	}
	if (strncmp(rule->cond, "size", 4) == 0 && rule->cond[5] != '\0') {
		size_t size;
		char *ep;

		if (rule->cond[4] == '\0' || rule->cond[5] == '\0' ||
		    rule->cond[6] == '\0') {
			return (0);
		}
		size = (size_t)strtoul(&rule->cond[6], &ep, 10);
		if (*ep != '\0') {
			syslog(LOG_ERR, "<%s>: Bad rule syntax: `%s'",
			    msg->rcpt_to, rule->cond);
			return (0);
		}
		return (rule->cond[4] == '<' ? (msg->text_len <= size) :
		                               (msg->text_len >= size));
	}
	return (0);
}

/*
 * Classify a message according to a ruleset. Rulesets are processed
 * recursively when the "&ref" action syntax is used.
 */
int
SF_MessageClassify(SF_Message *msg, SF_Ruleset *ruleset, int lvl)
{
	uid_t default_uid;
	gid_t default_gid;
	SF_Rule *rule;
	SF_Ruleset *pRuleset;
	
	if (LOCAL_GetDefaultRecipientUID(msg->rcpt_to,
	    &default_uid, &default_gid) == -1) {
		SF_SetError("Cannot get default UID for %s", msg->rcpt_to);
		return (-1);
	}
	TAILQ_FOREACH(rule, &ruleset->rules, rules) {
		uid_t uid = (rule->uid != 0) ? rule->uid : default_uid;
		gid_t gid = (rule->gid != 0) ? rule->gid : default_gid;
		
		if (getpwuid(uid) == NULL || getgrgid(gid) == NULL) {
			syslog(LOG_ERR, "<%s>: Bad UID %d:%d; ignored rule",
			    msg->rcpt_to, uid, gid);
			continue;
		}
		if (rule->flags & RULE_SMTP) {
			continue;				/* Ignore */
		}
		if ((!(rule->flags&RULE_NEGATE) && !SF_MatchRule(msg, rule)) ||
		    ( (rule->flags&RULE_NEGATE) &&  SF_MatchRule(msg, rule)))
			continue;

		Debug("Rule: [%s] %s (as %d:%d)", rule->cond, rule->insn,
		    (int)uid, (int)gid);

		if (rule->insn[0] == '|') {
			int rv;

			if ((rv = LOCAL_FeedToPipe(msg, &rule->insn[1],
			    rule->cond, uid, gid)) == -1)  {
				goto queue;
			} else if (rv == 99) {
				/* Abort on code 99 (qmail compatibility) */
				break;
			}
			continue;
		}
		switch (rule->insn[0]) {
		case '/':
			if (rule->insn[1] == 'd' &&
			    strcmp(rule->insn, "/dev/null") == 0) {
				/* Silently drop */
				return (0);
			}
			if (rule->insn[strlen(rule->insn)-1] == '/') {
				Debug("Delivering to maildir: %s", rule->insn);
				if (LOCAL_DeliverToMaildir(msg, rule->insn,
				    rule->cond, uid, gid) == -1)
					goto queue;
			} else {
				Debug("Delivering to mbox: %s", rule->insn);
				if (LOCAL_DeliverToMailbox(msg, rule->insn,
				    rule->cond, uid, gid) == -1)
					goto queue;
			}
			break;
		case '&':
			if (lvl >= RULE_MAX_RECURSION_LVL) {
				SF_SetError("Too many levels of recursion");
				return (-1);
			}
			if ((pRuleset = LOCAL_GetRulesetByName(&rule->insn[1]))
			    == NULL) {
				return (-1);
			}
			if (SF_MessageClassify(msg, pRuleset, lvl+1) == -1) {
				SF_RulesetFree(pRuleset);
				return (-1);
			}
			SF_RulesetFree(pRuleset);
			break;
		default:
			if (LOCAL_DeliverToAddress(msg, rule->insn, uid, gid)
			    == -1)
				goto queue;
		}
		Debug("Delivery successful: %s => %s", rule, rule->insn);
	}
	return (0);
queue:
	Debug("Reinjecting (%s)", SF_GetError());
	/* TODO Implement QUEUEING! (reinject for now) */
#if 0
	syslog(LOG_ERR, "Queueing message from <%s> (%s)", msg->mail_from,
	    SF_GetError());
	QueueMessage(msg);
	return (0);
#endif
	return (-1);
}

/*
 * Extract user/domain parts from an address.
 * Buffers must be able to hold ADDRESS_MAX characters.
 */
int
SF_ParseRecipientParts(const char *rcpt, char *user_part, char *domain_part)
{
	const char *c;
	int i;

	/* Parse the local part. */
	for (i = 0, c = &rcpt[0];
	     i < (ADDRESS_MAX-2) && *c != '\0' && *c != '@';
	     i++, c++) {
		user_part[i] = *c;
	}
	if (i == 0 || *c == '\0' || i == (ADDRESS_MAX-2)) {
		return (-1);
	}
	user_part[i] = '\0';
	c++;

	/* Parse the domain part. */
	for (i = 0;
	     i < (ADDRESS_MAX-2) && *c != '\0';
	     i++, c++) {
		domain_part[i] = *c;
	}
	if (i == 0 || i == (ADDRESS_MAX-2)) {
		return (-1);
	}
	domain_part[i] = '\0';
	return (0);
}

static char *
unescape_rule(char *rule)
{
	char hex[3];
	const char *sp;
	char *dst, *dp;
	unsigned int n;

	dp = dst = Malloc(strlen(rule)+2);
	hex[2] = '\0';
	for (sp = rule; *sp != '\0'; dp++, sp++) {
		if (sp[0] == '\\' && isxdigit(sp[1]) && isxdigit(sp[2])) {
			hex[0] = sp[1];
			hex[1] = sp[2];
			n = (unsigned int)strtoul(hex, NULL, 16);
			if (n == '\0') {
				*dp = '_';
			} else {
				*dp = n;
			}
			sp += 2;
		} else if (sp[0] == '+') {
			*dp = ' ';
		} else {
			*dp = sp[0];
		}
	}
	*dp = '\0';
	return (dst);
}

/* Parse a ruleset and return a ruleset structure. */
SF_Ruleset *
SF_RulesetParse(const char *key, char *data)
{
	char *data_dup , *pdata, *s;
	SF_Ruleset *rs;
	SF_Rule *r;

	if ((data_dup = strdup(data)) == NULL) {
		SF_SetError("Out of memory");
		return (NULL);
	}
	pdata = data_dup;

	rs = Malloc(sizeof(SF_Ruleset));
	strlcpy(rs->key, key, sizeof(rs->key));
	TAILQ_INIT(&rs->rules);
	while ((s = strsep(&pdata, ",")) != NULL) {
		char *cond = strsep(&s, "|");
		char *insn = strsep(&s, "|");
		char *uid = strsep(&s, "|");
		char *gid = strsep(&s, "|");

		if (cond == NULL || cond[0] == '\0' ||
		    insn == NULL || insn[0] == '\0') {
			continue;
		}
		for (; isspace(*cond); cond++) ;;
		for (; isspace(*insn); insn++) ;;
		
		r = Malloc(sizeof(SF_Rule));
		r->flags = 0;

		if (cond[0] == '!') {
			r->flags |= RULE_NEGATE;
			cond++;
		} else if (cond[0] == '>') {
			r->flags |= RULE_SMTP;
		}
		r->cond = unescape_rule(cond);
		r->insn = unescape_rule(insn);

		if (uid != NULL && uid[0] != '\0') {
			r->uid = (uid_t)atoi(uid);
		} else {
			r->uid = 0;
		}
		if (gid != NULL && gid[0] != '\0') {
			r->gid = (uid_t)atoi(gid);
		} else {
			r->gid = 0;
		}

		TAILQ_INSERT_TAIL(&rs->rules, r, rules);
	}

	free(data_dup);
	return (rs);
}

void
SF_RulesetFree(SF_Ruleset *rs)
{
	SF_Rule *r1, *r2;

	for (r1 = TAILQ_FIRST(&rs->rules);
	     r1 != TAILQ_END(&rs->rules);
	     r1 = r2) {
		r2 = TAILQ_NEXT(r1, rules);
		free(r1->cond);
		free(r1->insn);
		free(r1);
	}
}
